#include "bpp4.h"

#define BPP_COLOR 16

typedef struct {
  int r;
  int g;
  int b;
} PALETTE;

PALETTE PAL16[BPP_COLOR];
static int fore_index = 15;
static int back_index = 0;
int bpp4_color_index = 15;

#define PAL PAL16

int bpp4_nearest(int r,int g,int b)
{
  int dist=0xffffff;
  int n,d,i,R,G,B;
  for (n=0,i=0;n<BPP_COLOR;n++) {
    R = r - PAL[n].r; R*=R;
    G = g - PAL[n].g; G*=G;
    B = b - PAL[n].b; B*=B;
    d = R + G + B;
    if (d<dist) { i = n; dist = d; }
  }
  return i;
}

int bpp4_setcolor(int r,int g,int b)
{
  bpp4_color_index = bpp4_nearest(r,g,b);
  vga_setcolor(bpp4_color_index);
  return bpp4_color_index;
}

int bpp4_bitblt(int x,int y,int w,int h,char *bmp)
{
  int X,Y;
  int xp,yp;
  for (Y=0;Y<h;Y++) {
    yp = y + Y; xp = x;
    for (X=0;X<w;X++) {
      vga_setcolor((int)*bmp);
      vga_drawpixel(xp,yp);
      xp++; bmp++;
    }
  }
  return 1;
}

int bpp4_getbitmap(int x,int y,int w,int h,char *bmp)
{
  int X,Y;
  int xp,yp;
  for (Y=0;Y<h;Y++) {
    yp = y + Y; xp = x;
    for (X=0;X<w;X++) {
      *bmp = (char)vga_getpixel(xp,yp);
      xp++; bmp++;
    }
  }
  return 1;
}

int bpp4_scroll(int x,int y,int w,int h,int line)
{
  char * bmp;
  char * piece;
  int H;
  
  if (line==0) return 0;
  bmp = (char*)malloc(w*h*sizeof(char));
  bpp4_getbitmap(x,y,w,h,bmp);
  if (line>0) { /* scroll up */
    piece = &bmp[line*w]; H = h - line;
  } else { /* scroll down */
    piece = bmp; H = h + line; y-=line;
  }
  bpp4_bitblt(x,y,w,H,piece);
  free(bmp);
  return line;
}

int bpp4_fillbox(int x,int y,int w,int h)
{
  int x2 = x + w - 1;
  int n;
  for (n=0;n<h;n++,y++) vga_drawline(x,y,x2,y);
  return 1;
}

int bpp4_xorbox(int x,int y,int w,int h)
{
  unsigned char * bmp;
  unsigned char * p;
  int n = w * h;
  bmp = (unsigned char *)malloc(n);
  p = bmp;
  bpp4_getbitmap(x,y,w,h,bmp);
  while (n>0) { *p ^= 0xf; n--; p++; }
  bpp4_bitblt(x,y,w,h,bmp);
  free(bmp);
}

int bpp4_drawfont(int x,int y,int w,int h,unsigned char *bmp)
{
  char MASK[8]={0x80,0x40,0x20,0x10,0x8,0x4,0x2,0x1};
  int carry = w%8;
  int X,Y,i;
  int xp,yp;
  
  for (Y=0;Y<h;Y++) {
    xp = x; yp = y + Y;
    for (X=0,i=0;X<w;X++,xp++) {
      ((*bmp) & MASK[i]) ? vga_setcolor(fore_index):vga_setcolor(back_index);
      vga_drawpixel(xp,yp);
      if (i==7) bmp++;
      i++; i%=8;
    }
    if (carry) bmp++;
  }
}

int bpp4_setforeground(int r,int g,int b)
{
  set_swcolor(foreground,r,g,b);
  fore_index = bpp4_nearest(r,g,b);
  return fore_index;
}

int bpp4_setbackground(int r,int g,int b)
{
  set_swcolor(background,r,g,b);
  back_index = bpp4_nearest(r,g,b);
  return back_index;
}

extern int bpp4_miximage(unsigned char *dest,unsigned char *src,int w,int h);

unsigned char * bpp4_xpm(int *w,int *h,char **xpm)
{
  int x,y;
  int W,H;
  int colors;
  int n,i;
  int present;
  unsigned char * image;
  PALETTE *pal;
  int * MAP;
  char ** KEY;
  char * buff;
  unsigned char *p;

  if (sscanf(xpm[0],"%d %d %d %d",w,h,&colors,&present)!=4) return NULL;
  W = *w; H = *h;

  /* allocate memory */
  image = (unsigned char *)malloc(W*H*sizeof(char));

  /* color */
  pal = (PALETTE*)malloc(sizeof(PALETTE)*colors);
  MAP = (int*)malloc(sizeof(int)*(colors+1));
  memset(MAP,0,sizeof(int)*(colors+1));
  KEY = (char**)malloc(sizeof(char*)*colors);

  for (i=0,n=1;i<colors;i++,n++) {
    KEY[i]=(char*)malloc(sizeof(char)*(present+1));
    KEY[i][present]=0;
    memcpy(KEY[i],xpm[n],present);
    buff = &xpm[n][present];
    while (*buff&&*buff!='c') buff++;
    if (*buff=='c') buff++;
    if (*buff==' ') buff++;
    if (strcasecmp(buff,"None")==0) {
      memset(&pal[i],0,sizeof(PALETTE)); 
    } else {
      if (*buff=='#') buff++;
      #define GETID(v) (v>='0'&&v<='9') ? v-'0': ((v>='A'&&v<='Z') ? v-'A'+10: ((v>='a'&&v<='z') ? v-'a'+10:0))
      #define GETCOLOR(c) c=GETID(*buff); c<<=4; buff++; c|=GETID(*buff); buff++
      GETCOLOR(pal[i].r);
      GETCOLOR(pal[i].g);
      GETCOLOR(pal[i].b);
      #undef  GETCOLOR
      #undef  GETID
    }
    MAP[i] = bpp4_nearest(pal[i].r,pal[i].g,pal[i].b);
  }
  p = image; *p = 0;
  for (y=0;y<H;y++,n++) {
    for (x=0;x<W;x++) {
      for (i=0;i<colors;i++)
        if (strncmp(KEY[i],&xpm[n][x*present],present)==0) break;
      *p = (char)MAP[i];
      p++;
    }
  }
  
  free(pal);
  free(MAP);
  for (i=0;i<colors;i++) free(KEY[i]);
  free(KEY);
  return image;
}

int bpp4_mouseinit(char **xpm)
{
  mouse_bitmap = create_image_xpm(&mouse_width,&mouse_height,xpm);
  mouse_memsize = mouse_width*mouse_height*sizeof(char);
  mouse_screen = (unsigned char *)malloc(mouse_memsize);
  mouse_paint  = (unsigned char *)malloc(mouse_memsize);
  
  return 1;
}

int bpp4_mouserecover(void)
{
  if (!mouse_enable) return 0;
  if (mouse_painted) {
    bitblt(mouse_x,mouse_y,mouse_width,mouse_height,mouse_screen);
  }
  mouse_painted = 0;
  mice_update  = 0;
  return 1;
}

int bpp4_mousemove(int x,int y)
{
  if (!mouse_enable) return 0;
  if (mouse_painted) {
    bitblt(mouse_x,mouse_y,mouse_width,mouse_height,mouse_screen);
  }
  getbitmap(x,y,mouse_width,mouse_height,mouse_screen);
  memcpy(mouse_paint,mouse_screen,mouse_memsize);
  miximage(mouse_paint,mouse_bitmap,mouse_width,mouse_height);
  mouse_x = x;
  mouse_y = y;
  bitblt(x,y,mouse_width,mouse_height,mouse_paint);
  mouse_painted = 1;
  mice_update  = 0;
  return 1;
}

unsigned char * bpp4_createbuffer(int w,int h)
{
  return (unsigned char *)malloc(w*h*sizeof(char));
}

int  bpp4_init(void)
{
  int n;
  /* Get Palette */
  for (n=0;n<BPP_COLOR;n++)  {
    vga_getpalette(n,&PAL[n].r,&PAL[n].g,&PAL[n].b);
    PAL[n].r <<= 2;
    PAL[n].g <<= 2;
    PAL[n].b <<= 2;
  }
  drawpixel        = vga_drawpixel;
  getpixel         = vga_getpixel;
  drawline         = vga_drawline;
  setcolor         = bpp4_setcolor;
  bitblt           = bpp4_bitblt;
  getbitmap        = bpp4_getbitmap;
  miximage         = bpp4_miximage;
  scroll           = bpp4_scroll;
  fillbox          = bpp4_fillbox;
  xorbox           = bpp4_xorbox;
  drawfont         = bpp4_drawfont;
  setforeground    = bpp4_setforeground;
  setbackground    = bpp4_setbackground;
  create_image_xpm = bpp4_xpm;
  mice_grinit      = bpp4_mouseinit;
  create_buffer    = bpp4_createbuffer;
  
  return 1;
}


